'''
    Funções responsáveis pela extração de dados da base Siapa.
'''

from siapatools import SIAPA

CANCELADO_PARC_MANUAL = '20'


def busca_debitos_cancelados():
    '''
    Retorna os débitos cancelados por parcelamento manual da base Siapa.
    '''
    siapa = SIAPA()
    siapa.minAnoDtBase = '1964'
    tabela = siapa.pega_debitos(filtro_situacao=[CANCELADO_PARC_MANUAL],
                                colunas=['nu_debito'])
    debitos = set(tabela['nu_debito'])
    qtd_debitos = len(debitos)
    msg = f'{qtd_debitos} débitos cancelados por parcelamento manual.'
    #print(msg)
    siapa.con.close()
    return debitos


def busca_uf_rip():
    siapa = SIAPA()
    imoveis = siapa.imoveis
    uf_rip = imoveis.merge(siapa.UF, how='left', left_on='ed_numero_uf',
                           right_on='nu_uf')
    colunas = ['nu_rip', 'sg_uf']
    uf_rip = uf_rip[colunas]
    uf_rip = uf_rip.rename(columns={'sg_uf': 'uf_rip'})
    erros = len(uf_rip[uf_rip['uf_rip'].isnull()])
    assert erros == 0
    return uf_rip



'''
    CLI para o script de cancelamento de débitos.
'''
import os
import sys

import fire

from .cancelador import Cancelador
from .config import config_siapa, config_spunet
#from .valida import main as Validador
from .valida import main_multprocess as Validador
from .__init__ import __version__

def valida_config(config):
    for k, value in config.items():
        if value == '':
            msg = f'{k} inválido. Verifique a configuração no arquivo config.py do pacote.'
            print(msg)
            return False
    return True

def cli(validador=False):
    os.system('cls')
    print(f'CANCELA DÉBITOS PARCELAMENTO v{__version__}')
    print(f'===================================\n')

    if not all([valida_config(config_siapa), valida_config(config_spunet)]):
        sys.exit(1)

    if validador:
        Validador(config_siapa['cpf'], config_siapa['senha'])
    else:
        Cancelador(config_siapa['cpf'], config_siapa['senha'])


def main():
    fire.Fire(cli)



'''
    Busca os débitos parcelados no banco de dados do SPUnet e faz o
    cancelamento no SIAPA.
'''

from collections import namedtuple
import logging
import os
import sys

import pandas as pd
import psycopg2
from selenium.common.exceptions import TimeoutException, NoSuchElementException
from tqdm import tqdm
from siapa_robo import Siapa

from .bancodedados import Parcelamento
from .dadossiapa import busca_debitos_cancelados, busca_uf_rip


__all__ = ['ClasseDebito', 'busca_uf_rip', 'marca_uf_debitos',
           'busca_debitos_para_baixa', 'Cancelador', 'exclui_debitos_baixados']


ERROS_SIAPA = {'5045': '5045 - SITUACAO DO DEBITO NAO PERMITE CANCELAMENTO',
               '5022': '5022 - DEBITO INFORMADO TEM RECEITA ORDINARIA'}


class ClasseDebito():
    ORDINARIO = 'Ordinário'
    EXTRAORDINARIO = 'Extraordinário'

    def __init__(self):
        pass

    @property
    def classes(self):
        '''
            Classes de receita do SIAPA.
        '''
        nomes_classes = [self.ORDINARIO, self.EXTRAORDINARIO]
        classes_debitos = namedtuple('classes_debitos', nomes_classes)
        return classes_debitos(('2073', '2090'), ('2081', '9136'))

    def classifica(self, co_receita):
        ''' Classifica um código de receita como extraordinário ou
            ordinário.
        '''
        for codigos, classe_debito in zip(self.classes, self.classes._fields):
            if co_receita in codigos:
                classificacao = classe_debito
                break
        else:
            raise ValueError('Código desconhecido {}'.format(co_receita))

        return classificacao


def seleciona_menu(siapa, classe_debito):
    #menu = {ClasseReceita.ordinaria: 3, ClasseReceita.extraordinaria: 6}
    menu = {ClasseDebito.ORDINARIO: 3, ClasseDebito.EXTRAORDINARIO: 6}
    siapa.navigate_menu(8, 1, menu[classe_debito])
    titulo = siapa.driver.find_element_by_tag_name('p').text
    assert classe_debito in titulo
    return None


def marca_uf_debitos(dataframe_debitos):
    df = dataframe_debitos.copy()
    df['nu_rip'] = df['nu_rip'].map(lambda x: x[:11])
    uf_rip = busca_uf_rip()
    marcados = df.merge(uf_rip, how='left', on='nu_rip')
    assert not marcados['uf_rip'].isnull().any()
    return marcados


def busca_debitos_para_baixa():
    '''
    Retorna dataframe contendo os débitos referentes às operações de
    parcelamento já deferidas.
    '''
    p = Parcelamento()
    deferidos = p.filtra_parcelamentos(situacao='Deferido')
    qtd_deferidos = len(deferidos)
    #print(f'Encontrados {qtd_deferidos} parcelamentos deferidos.')

    debitos_deferidos = p.tb_debitos_parcelados.merge(
            deferidos, how='inner', on='id_parcelamento')
    #print(f'Encontrados {len(debitos_deferidos)} débitos referentes a parcelamentos deferidos.')


    lista_final = marca_uf_debitos(debitos_deferidos)

    lista_final['classe_receita'] = lista_final['id_tiporeceita'].map(ClasseDebito().classifica)

    cols = ['nu_debito', 'nu_rip', 'uf_rip', 'classe_receita', 'id_parcelamento']
    lista_final = lista_final[cols].sort_values(by=['uf_rip', 'id_parcelamento'])
    return lista_final


def exclui_debitos_baixados(df):
    '''
    Exclui da tabela de débitos para baixa os débitos que já estão cancelados
    na base Siapa.
    '''
    cancelados = busca_debitos_cancelados()
    nao_baixados = df[~df['nu_debito'].isin(cancelados)]
    return nao_baixados


def cancela_debito(siapa, nu_debito):
    try:
        form = siapa.formulario()
    except TimeoutException:
        logging.error('Não foi possível carregar o formulário. Saindo do '
                      'programa.')
        raise TimeoutException

    campos = form.fields
    campos['TextDeb'].clear()
    campos['TextDeb'].send_keys(nu_debito)

    campos['TextMot'].options[7].click()

    campos['TextProc'].clear()
    nu_processo_sei = '04905002144201714'
    campos['TextProc'].send_keys(nu_processo_sei)

    form.submit()

    #verifica se houve erro. Informa o erro e passa para o próximo débito.
    try:
        erro_siapa = siapa.driver.find_element_by_tag_name('li')
        if erro_siapa:
            msg_erro = erro_siapa.text
            botao_voltar = 'Voltar'
            siapa.driver.find_element_by_name(botao_voltar).click()
            return msg_erro[:4]
    except:
        BOTAO_CONFIRMAR = 'Desistir'  # botão possui o nome errado no Siapa
        siapa.driver.find_element_by_name(BOTAO_CONFIRMAR).click()

        botao_voltar = 'Retorna'
        siapa.driver.find_element_by_name(botao_voltar).click()

        return True


def crash_recover(siapa, cpf, senha, UF, tipo_debito):
    ''' Reabre o Siapa e retorna para a tela de alteração de débito '''
    print('\nAtivando o Crash Recover.')
    siapa.driver.close()
    siapa.entrar(numCpf=cpf, txtSenha=senha)
    siapa.trocar_GRPU(UF)
    seleciona_menu(siapa, tipo_debito)


def busca_debitos_UF(sigla_UF, dataframe):
    ''' Busca os débitos ordinário e extraordinarios da UF selecionada '''

    debitos_UF = dataframe.query(f'uf_rip == "{sigla_UF}"')

    query_receita_ord = f'classe_receita == "{ClasseDebito.ORDINARIO}"'
    ordinarios = debitos_UF.query(query_receita_ord)

    query_receita_ext = f'classe_receita == "{ClasseDebito.EXTRAORDINARIO}"'
    extraordinarios = debitos_UF.query(query_receita_ext)
    return ordinarios, extraordinarios


def cancela_lista_debitos(siapa, df, classe_debito, UF, pbar):
    ''' Realiza o cancelamento de todos os débitos de uma determinada classe
    contidos em um dataframe '''
    qtd_debitos = len(df)
    if qtd_debitos > 0:
        msg = f'{UF} - Iniciando a baixa de {qtd_debitos} débitos {classe_debito}.'
        logging.info(msg)

        try:
            seleciona_menu(siapa, classe_debito)
        except TimeoutException:
            crash_recover(siapa, nu_cpf, txt_senha, UF, classe_debito)

        for row in df.iterrows():
            dados = row[1]
            try:
                status = cancela_debito(siapa, dados.nu_debito)
            except (NoSuchElementException, TimeoutException):
                logging.info(f'Erro. Ativado o crash recover')
                crash_recover(siapa, nu_cpf, txt_senha, UF, classe_debito)

                logging.info(f'Efetuando nova tentativa')
                status = cancela_debito(siapa, dados.nu_debito)

            if status in ERROS_SIAPA:
                msg = f'{UF} - {dados.nu_debito} - {ERROS_SIAPA[status]}'
                logging.info(msg)
            elif status is True:
                msg = f'{UF} - {dados.nu_debito} - Transação efetuada com sucesso.'
                logging.info(msg)
            else:
                msg = f'{UF} - {dados.nu_debito} - Status inesperado {status}.'
                logging.error(msg)
                raise EnvironmentError(msg)

            pbar.update(1)

        logging.info(f'{UF} - Todos os débitos ordinários processados com sucesso.')
        botao_menu = 'Voltar'
        siapa.driver.find_element_by_name(botao_menu).click()
    else:
        logging.info(f'{UF} - UF não possui débitos {classe_debito} para cancelamento.')
        return


def Cancelador(cpf, senha):
    ''' Função principal da rotina '''
    global nu_cpf
    global txt_senha

    nu_cpf = cpf
    txt_senha = senha

    logging.basicConfig(filename='log.txt', level=logging.INFO,
                        format='%(asctime)s:%(levelname)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    debitos = busca_debitos_para_baixa()
    print(' - Excluindo os débitos já cancelados na base Siapa.')
    debitos = exclui_debitos_baixados(debitos)

    qtd_debitos = len(debitos)
    msg_qtd_debitos = f' - Encontrados {qtd_debitos} débitos para cancelamento.'
    print(msg_qtd_debitos)
    logging.info(msg_qtd_debitos)

    spu_com_operacoes = debitos['uf_rip'].unique()

    msg_inicializa = ' - Inicializando o Siapa.'
    logging.info(msg_inicializa)
    print(msg_inicializa)
    siapa = Siapa(headless=True)

    msg_login = f' - Efetuando login no Siapa como {cpf}.'
    logging.info(msg_login)
    print(msg_login)
    siapa.entrar(numCpf=nu_cpf, txtSenha=txt_senha)

    for UF in spu_com_operacoes:
        logging.info(f'{UF} - INICIANDO O PROCESSAMENTO')

        siapa.trocar_GRPU(UF)
        logging.info(f'{UF} - GRPU alterada para com sucesso.')

        ordinarios_UF, extraordinarios_UF = busca_debitos_UF(UF, dataframe=debitos)

        # Remove os débitos já cancelados na base Siapa
        qtd_ordinarios = len(ordinarios_UF)
        qtd_extraordinarios = len(extraordinarios_UF)
        logging.info(f'{UF} - Ordinários: {qtd_ordinarios} Extraordinários: {qtd_extraordinarios}')


        with tqdm(total=qtd_ordinarios + qtd_extraordinarios) as pbar:
            cancela_lista_debitos(
                siapa, ordinarios_UF, ClasseDebito.ORDINARIO, UF, pbar)
            logging.info(f'Retornou ao menu principal.')

            cancela_lista_debitos(
                siapa, extraordinarios_UF, ClasseDebito.EXTRAORDINARIO,
                UF, pbar)
            logging.info(f'Retornou ao menu principal.')

        logging.info(f'{UF} - TODOS DÉBITOS PROCESSADOS.')

    msg_fim = f' - Cancelamento realizado com sucesso. Todas UF processadas.'
    logging.info(msg_fim)
    print(msg_fim)

    return True



'''
    Acessa o banco de dados do SPUNET e busca as informações das operações
    de parcelamento.
'''

import pandas as pd
import psycopg2

from .config import config_spunet


class Parcelamento():
    '''
    Acessa os dados das operações de parcelamento no SPUNET.
    '''
    def __init__(self):
        con_args = {'host': '10.209.9.199',
                    'dbname': 'receitas',
                    'user': config_spunet['user'],
                    'password': config_spunet['password']}

        self.con = psycopg2.connect(**con_args)

    def get_table(self, schema_table):
        '''
        Retorna uma dataframe contendo uma tabela do banco de dados

        Parâmetros:
          schema_table(str): Caminho para a tabela. Ex: "siapa.tb_debitos"
        '''

        sql = "SELECT * FROM {};".format(schema_table)
        table = pd.read_sql_query(sql, con=self.con)
        return table

    @property
    def tb_parcelamento(self):
        ''' Retorna a tabela com as operações de parcelamento '''

        parc = self.get_table('siapa.tb_novo_parcelamento')
        parc = parc.merge(self.tb_situacao, how='left', left_on='id_situacao',
                          right_on='id_situacao_parcelamento')
        del parc['id_situacao_parcelamento']
        return parc

    @property
    def qtd_parcelamentos(self):
        ''' Quantidade total de parcelamentos '''
        return len(self.tb_parcelamento)

    @property
    def tb_debitos_parcelados(self):
        ''' Retorna a tabela de débitos parcelados '''
        debitos = self.get_table('siapa.tb_debito_parcelado')
        return debitos

    @property
    def qtd_debitos(self):
        ''' Quantidade total de débitos parcelados '''
        return len(self.tb_debitos_parcelados)

    @property
    def tb_situacao(self):
        ''' Tabela de domínio com as situações possível para operações
        de parcelamento '''
        return self.get_table('siapa.tb_situacao_parcelamento')

    @property
    def resumo_por_situacao(self):
        '''
            Resumo das operações de parcelamento consolidado por situação.
        '''
        resumo = self.tb_parcelamento.groupby('ds_situacao_parcelamento').count()['id_parcelamento']
        resumo = resumo.reset_index().rename(columns={'id_parcelamento': 'Quantidade'})
        return resumo

    @property
    def resumo_por_uf(self):
        ''' Operações de parcelamento consolidadas por UF '''

        cols = ['sg_uf', 'ds_situacao_parcelamento']
        resumo = self.tb_parcelamento.groupby(cols)['id_parcelamento'].count()
        resumo = resumo.reset_index().rename(columns={'id_parcelamento': 'Quantidade'})
        resumo = resumo.sort_values(by=['sg_uf', 'Quantidade'])
        return resumo

    def filtra_parcelamentos(self, situacao=None, sigla_uf=None):
        ''' Retorna as operações de parcelamento com a situação selecionada '''

        filtro_situacao = f'ds_situacao_parcelamento == "{situacao}"'
        filtro_uf = f'sg_uf == "{sigla_uf}"'
        params = filter(lambda x: "None" not in x, [filtro_situacao, filtro_uf])
        return self.tb_parcelamento.query(' & '.join(list(params)))

    def busca_debitos_parcelamento(self, id_parcelamento):
        ''' id_parcelamento (int) '''

        if isinstance(id_parcelamento, int) is False:
            raise ValueError('id_parcelamento deve ser int')

        query = f'id_parcelamento == {id_parcelamento}'
        debitos = self.tb_debitos_parcelados.query(query)
        return debitos



from setuptools import setup
from canceladebitos.__init__ import __version__

setup(
    name="canceladebitos",
    packages=["canceladebitos"],
    entry_points={
        "console_scripts": ['canceladebitos = canceladebitos.__main__:main']
        },
    version=__version__,
    description="Cancelamentos dos débitos renegociados em operações de parcelamento.",
    author="Rafael Alves Ribeiro",
    author_email="rafael.alves.ribeiro@gmail.com",
    install_requires=[
        "codecov",
        "fire",
        "numpy",
        "pandas",
        "psycopg2",
        "pytest",
        "selenium",
        "siapatools",
        "siapa_robo",
        "tqdm",        
    ],
    )



import time
import sys
from functools import partial
from multiprocessing import Pool, Process, Value, Manager

import numpy as np
import pandas as pd
from tqdm import tqdm

from selenium.common.exceptions import NoSuchFrameException
from siapa_robo import Siapa

from .cancelador import *


def login_siapa(siapa, nu_cpf, senha):
    siapa.entrar(numCpf=nu_cpf, txtSenha=senha)
    return siapa

def acessa_consulta(siapa):
    MENU_CONSULTA_DEBITOS = (8, 6, 1)
    siapa.navigate_menu(*MENU_CONSULTA_DEBITOS)


def carrega_debitos():
    debitos = busca_debitos_para_baixa()
    debitos['nu_debito'] = debitos['nu_debito'].map(lambda x: str(x).zfill(8))
    debitos = debitos[['nu_rip', 'nu_debito']]
    return debitos


def envia_formulario(siapa, nu_rip):
    form = siapa.formulario()
    txt_RIP = form.fields['TextRip']
    txt_RIP.clear()
    txt_RIP.send_keys(nu_rip)
    botao_consultar = siapa.driver.find_element_by_name('Consultar')
    botao_consultar.click()


def clean_nu_debito(nu_debito):
    if pd.isnull(nu_debito):
        return None
    string = str(nu_debito)
    remove_decimal = string.replace('.0', '')
    numero = remove_decimal.zfill(8)
    return numero


def historico_debitos_cancelados(siapa):
    siapa.driver.switch_to_default_content()
    siapa.driver.switch_to_frame('Principal')
    siapa.driver.switch_to_frame('Dados_Principais')
    tabela = siapa.tables()[1]
    tabela['nu_debito'] = tabela['Nro Débito'].apply(clean_nu_debito)
    tabela['nu_debito'] = tabela['nu_debito'].fillna(method='ffill')
    tabela.rename(columns={'Valor na Data Base Calculo': 'situacao'},
                  inplace=True)
    CANCELADOS_PARC_MANUAL = 'CANC.PARC.MAN.'
    hist_cancelados = tabela.query(f'situacao == "{CANCELADOS_PARC_MANUAL}"')
    siapa.driver.switch_to_default_content()
    return hist_cancelados[['nu_debito', 'situacao']]

def valida_debitos():
    dados = tabelas[1]


def valida_rip(siapa, debitos, nu_rip):
    envia_formulario(siapa, nu_rip)
    try:
        historico = historico_debitos_cancelados(siapa)
    except (ValueError, NoSuchFrameException):
        ERRO_6904 = '6904 - RETORNO ANORMAL DO ACESSO AO TS'
        if ERRO_6904 in siapa.driver.page_source:
            #print(ERRO_6904)
            siapa.driver.find_element_by_name('Voltar').click()
            envia_formulario(siapa, nu_rip)
            historico = historico_debitos_cancelados(siapa)


    debitos_rip = debitos.query(f'nu_rip == "{nu_rip}"')['nu_debito'].values

    for dbt in debitos_rip:
        if not dbt in historico['nu_debito'].values:
            raise AssertionError(f'RIP: {nu_rip} - Débito: {dbt} não cancelado.')
            sys.exit(1)
    siapa.driver.back()


def crash_recover(siapa, nu_cpf, senha, debitos, nu_rip):
    siapa.driver.quit()
    login_siapa(siapa, nu_cpf, senha)
    acessa_consulta(siapa)
    valida_rip(siapa, debitos, nu_rip)

def debitos_para_baixar():
    debitos = carrega_debitos()
    debitos = exclui_debitos_baixados(debitos)
    debitos.reset_index(inplace=True)
    return debitos


def fatia_dataframe(df, n):
    ''' Divide o dataframe em n partes e retorna uma lista de n dataframes '''
    tamanho_fatia = len(df) // n
    fatias = [i for g, i in df.groupby(np.arange(len(df)) // tamanho_fatia)]
    return fatias


def login_siapa_abre_consulta(nu_cpf, senha):
    try:
        siapa = Siapa(headless=True)
        login_siapa(siapa, nu_cpf, senha)
        acessa_consulta(siapa)
    except(TimeoutException) as err:
        siapa.driver.quit()
        siapa = login_siapa_abre_consulta(nu_cpf, senha)
    else:
        return siapa


def processo(nu_cpf, senha, counter, debitos):
    #print('Iniciando processo.')
    siapa = login_siapa_abre_consulta(nu_cpf, senha)

    rips = debitos['nu_rip'].unique()

    #with tqdm(total=qtd_rips, unit='RIP') as pbar:
    for nu_rip in rips:
        try:
            valida_rip(siapa, debitos, nu_rip)
        except (ValueError, NoSuchFrameException, UnboundLocalError):
            while True:
                try:
                    crash_recover(siapa, nu_cpf, senha, debitos, nu_rip)
                except:
                    pass
                else:
                    break

        with counter.get_lock():
            counter.value += 1

    siapa.driver.quit()
    return


def pbar(total, counter):
    with tqdm(total=total) as barra:
        while counter.value <= total:
            barra.update(counter.value - barra.n)
    return

def main_multprocess(nu_cpf, senha):
    print(' - Iniciando validação do processo de cancelamento dos '
          'débitos parcelados.')
    debitos = debitos_para_baixar()
    N_PROCESSOS = 4
    chunks = fatia_dataframe(debitos, N_PROCESSOS)
    qtd_rips = len(debitos['nu_rip'].unique())
    print(f' - {qtd_rips} imóveis contendo {len(debitos)} débitos selecionados '
          'para validação.')

    counter = Value('i', 0)

    p1 = Process(target=processo, args=(nu_cpf, senha, counter, chunks[0]))
    p2 = Process(target=processo, args=(nu_cpf, senha, counter, chunks[1]))
    p3 = Process(target=processo, args=(nu_cpf, senha, counter, chunks[2]))
    p4 = Process(target=processo, args=(nu_cpf, senha, counter, chunks[3]))
    bar = Process(target=pbar, args=(qtd_rips, counter))

    p1.start()
    p2.start()
    p3.start()
    p4.start()
    bar.start()

    p1.join()
    p2.join()
    p3.join()
    p4.join()
    bar.join()

    print(' - Validação realizada com sucesso. Todos os débitos estão cancelados.')
    sys.exit(0)
    return 0


def main(nu_cpf, senha):
    print(' - Iniciando validação do cancelamento dos débitos parcelados.')
    siapa = Siapa(headless=True)
    print(f' - Efetuando login no Siapa como {nu_cpf}.')
    login_siapa(siapa, nu_cpf, senha)
    acessa_consulta(siapa)

    print(f' - Buscando lista de débitos parcelados para validação.')
    debitos = carrega_debitos()
    print(f' - Excluindo os débitos já cancelados na base Siapa.')
    debitos = exclui_debitos_baixados(debitos)

    rips = debitos['nu_rip'].unique()
    print(f' - Processando {len(debitos)} débitos encontrados em {len(rips)} imóveis.')

    with tqdm(total=len(rips), unit='RIP') as pbar:
        for nu_rip in rips:
            try:
                valida_rip(siapa, debitos, nu_rip)
            except (NoSuchFrameException, ValueError) as err:
                print('\n', err, '- Ativando o Crash Recover')
                crash_recover(siapa, nu_cpf, senha, debitos, nu_rip)
            finally:
                pbar.update(1)

    siapa.driver.quit()
    print(' - Validação realizada com sucesso. Todos os débitos estão cancelados.')
    return 0
